from classic.migrations.backends.contrib.pymssql import PyMSSQLBackend
from classic.migrations.backends.base import DatabaseBackend
from classic.migrations.backends.base import get_backend_class
from classic.migrations.backends.core import MySQLBackend
from classic.migrations.backends.core import SQLiteBackend
from classic.migrations.backends.core import PostgresqlBackend
from classic.migrations.backends.core import PostgresqlPsycopgBackend

__all__ = [
    "PyMSSQLBackend",
    "DatabaseBackend",
    "get_backend_class",
    "MySQLBackend",
    "SQLiteBackend",
    "PostgresqlBackend",
    "PostgresqlPsycopgBackend",
]



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from classic.migrations.backends.base import DatabaseBackend


class ODBCBackend(DatabaseBackend):
    driver_module = "pyodbc"

    def connect(self, dburi):
        args = [
            ("UID", dburi.username),
            ("PWD", dburi.password),
            ("ServerName", dburi.hostname),
            ("Port", dburi.port),
            ("Database", dburi.database),
        ]
        args.extend(dburi.args.items())
        s = ";".join("{}={}".format(k, v) for k, v in args if v is not None)
        return self.driver.connect(s)



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import count
import configparser
import os
import random
import re
import string
import sys
import unicodedata

try:
    import termios

except ImportError:
    try:
        from msvcrt import getwch as getch  # type: ignore
    except ImportError:
        # some non Windows environments don't have termios (google cloud)
        # running yoyo through the python sdk should not require `getch`
        pass

else:

    def getch():
        """
        Read a single character without echoing to the console and without
        having to wait for a newline.
        """
        fd = sys.stdin.fileno()
        saved_attributes = termios.tcgetattr(fd)
        try:
            attributes = termios.tcgetattr(fd)  # get a fresh copy!
            attributes[3] = attributes[3] & ~(termios.ICANON | termios.ECHO)  # type: ignore  # noqa: E501
            attributes[6][termios.VMIN] = 1  # type: ignore
            attributes[6][termios.VTIME] = 0  # type: ignore
            termios.tcsetattr(fd, termios.TCSANOW, attributes)

            a = sys.stdin.read(1)
        finally:
            # be sure to reset the attributes no matter what!
            termios.tcsetattr(fd, termios.TCSANOW, saved_attributes)
        return a


def prompt(prompt, options):
    """
    Display the given prompt and list of options and return the user selection.
    """

    while True:
        sys.stdout.write("%s [%s]: " % (prompt, options))
        sys.stdout.flush()
        ch = getch()
        if ch == os.linesep:
            ch = (
                [o.lower() for o in options if "A" <= o <= "Z"] + list(options.lower())
            )[0]
        print(ch)
        if ch.lower() not in options.lower():
            print("Invalid response, please try again!")
        else:
            break

    return ch.lower()


def confirm(s, default=None):
    options = "yn"
    if default:
        default = default.lower()
        if default == "y":
            options = "Yn"
        elif default == "n":
            options = "yN"
    return prompt(s, options) == "y"


def plural(quantity, one, plural):
    """
    >>> plural(1, '%d dead frog', '%d dead frogs')
    '1 dead frog'
    >>> plural(2, '%d dead frog', '%d dead frogs')
    '2 dead frogs'
    """
    if quantity == 1:
        return one.replace("%d", "%d" % quantity)
    return plural.replace("%d", "%d" % quantity)


def get_editor(settings):
    """
    Return the user's preferred visual editor
    """
    return settings.EDITOR


def get_random_string(length, chars=(string.ascii_letters + string.digits)):
    """
    Return a random string of ``length`` characters
    """
    rng = random.SystemRandom()
    return "".join(rng.choice(chars) for i in range(length))


def change_param_style(target_style, sql, bind_parameters):
    """
    :param target_style: A DBAPI paramstyle value (eg 'qmark', 'format', etc)
    :param sql: An SQL str
    :bind_parameters: A dict of bind parameters for the query

    :return: tuple of `(sql, bind_parameters)`. ``sql`` will be rewritten with
             the target paramstyle; ``bind_parameters`` will be a tuple or
             dict as required.
    """
    if target_style == "named":
        return sql, bind_parameters
    positional = target_style in {"qmark", "numeric", "format"}
    if not bind_parameters:
        return (sql, (tuple() if positional else {}))

    _param_counter = count(1)

    def param_gen_qmark(name):
        return "?"

    def param_gen_numeric(name):
        return f":{next(_param_counter)}"

    def param_gen_format(name):
        return "%s"

    def param_gen_pyformat(name):
        return f"%({name})s"

    param_gen = {
        "qmark": param_gen_qmark,
        "numeric": param_gen_numeric,
        "format": param_gen_format,
        "pyformat": param_gen_pyformat,
    }[target_style]

    pattern = re.compile(
        # Don't match if preceded by backslash (an escape)
        # or ':' (an SQL cast, eg '::INT')
        r"(?<![:\\])"
        # one of the given bind_parameters
        r":("
        + "|".join(re.escape(k) for k in bind_parameters)
        + r")"
        # followed by a non-word char, or end of string
        r"(?=\W|$)"
    )

    transformed_sql = pattern.sub(lambda match: param_gen(match.group(1)), sql)
    if positional:
        positional_params = []
        for match in pattern.finditer(sql):
            param_name = match.group(1)
            positional_params.append(bind_parameters[param_name])
        return transformed_sql, tuple(positional_params)
    return transformed_sql, bind_parameters


def unidecode(s: str) -> str:
    """
    Return ``s`` with unicode diacritics removed.
    """
    combining = unicodedata.combining
    return "".join(c for c in unicodedata.normalize("NFD", s) if not combining(c))



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from datetime import datetime

from classic.migrations import exceptions
from classic.migrations.backends.core.postgresql import PostgresqlBackend


class RedshiftBackend(PostgresqlBackend):
    def list_tables(self):
        current_schema = self.execute("SELECT current_schema()").fetchone()[0]
        return super(PostgresqlBackend, self).list_tables(schema=current_schema)

    # Redshift does not support ROLLBACK TO SAVEPOINT
    def savepoint(self, id):
        pass

    def savepoint_release(self, id):
        pass

    def savepoint_rollback(self, id):
        self.rollback()

    # Redshift does not enforce primary and unique keys
    def _insert_lock_row(self, pid, timeout, poll_interval=0.5):
        poll_interval = min(poll_interval, timeout)
        started = time.time()
        while True:
            with self.transaction():
                # prevents isolation violation errors
                self.execute("LOCK {}".format(self.lock_table_quoted))
                cursor = self.execute(
                    "SELECT pid FROM {}".format(self.lock_table_quoted)
                )
                row = cursor.fetchone()
                if not row:
                    self.execute(
                        "INSERT INTO {} (locked, ctime, pid) "
                        "VALUES (1, :when, :pid)".format(self.lock_table_quoted),
                        {"when": datetime.utcnow(), "pid": pid},
                    )
                    return
                elif timeout and time.time() > started + timeout:
                    raise exceptions.LockTimeout(
                        "Process {} has locked this database "
                        "(run migrations break-lock to remove this lock)".format(row[0])
                    )
                else:
                    time.sleep(poll_interval)



from setuptools import setup, find_namespace_packages

setup(
    name='classic-migrations',
    version='0.0.8',
    description='Database migrations with SQL',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    author='Kazakovtsev Nikolay',
    author_email='acidos@yandex.ru',
    url='https://github.com/variasov/classic-migrations',
    keywords='migrations migrate database alembic south',
    license='Apache',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Database',
        'Topic :: Database :: Front-Ends',
        'Topic :: Software Development',
        'Topic :: Software Development :: Version Control',
        'Programming Language :: Python :: 3',
    ],
    packages=find_namespace_packages(where='sources'),
    python_requires='>=3.10',
    install_requires=[
        'sqlparse',
        'tabulate',
        'importlib_metadata>=3.6.0',
        'pydantic-settings~=2.6',
        'pydantic~=2.10.2',
    ],
    package_dir={'': 'sources'},
    extras_require={
        'mysql': ['PyMySQL'],
        'postgres': ['psycopg2-binary==2.9.9'],
        'pyodbc': ['pyodbc'],
        'pymssql': ['pymssql==2.3.1'],
    },
    entry_points={
        'console_scripts': [
            'migrations = classic.migrations.scripts.main:main',
        ],
        'migrations.backends': [
            'pymssql = classic.migrations.backends.contrib.pymssql:PyMSSQLBackend',
            'odbc = classic.migrations.backends.contrib.odbc:ODBCBackend',
            'oracle = classic.migrations.backends.contrib.oracle:OracleBackend',
            'postgres = classic.migrations.backends.core.postgresql:PostgresqlBackend',
            'postgresql = classic.migrations.backends.core.postgresql:PostgresqlBackend',
            'psql = classic.migrations.backends.core.postgresql:PostgresqlBackend',
            'postgresql+psycopg = classic.migrations.backends.core.postgresql:PostgresqlPsycopgBackend',
            'mysql = classic.migrations.backends.core.mysql:MySQLBackend',
            'mysql+mysqldb = classic.migrations.backends.core.mysql:MySQLdbBackend',
            'sqlite = classic.migrations.backends.core.sqlite3:SQLiteBackend',
            'snowflake = classic.migrations.backends.contrib.snowflake:SnowflakeBackend',
            'redshift = classic.migrations.backends.contrib.redshift:RedshiftBackend',
        ],
    },
)



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from classic.migrations.backends.base import DatabaseBackend


class OracleBackend(DatabaseBackend):

    driver_module = "cx_Oracle"
    list_tables_sql = "SELECT table_name FROM all_tables WHERE owner=user"

    def begin(self):
        """Oracle is always in a transaction, and has no "BEGIN" statement."""
        self._in_transaction = True

    def connect(self, dburi):
        kwargs = dburi.args
        if dburi.username is not None:
            kwargs["user"] = dburi.username
        if dburi.password is not None:
            kwargs["password"] = dburi.password
        # Oracle combines the hostname, port and database into a single DSN.
        # The DSN can also be a "net service name"
        kwargs["dsn"] = ""
        if dburi.hostname is not None:
            kwargs["dsn"] = dburi.hostname
        if dburi.port is not None:
            kwargs["dsn"] += ":{0}".format(dburi.port)
        if dburi.database is not None:
            if kwargs["dsn"]:
                kwargs["dsn"] += "/{0}".format(dburi.database)
            else:
                kwargs["dsn"] = dburi.database

        return self.driver.connect(**kwargs)



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from classic.migrations.backends.base import DatabaseBackend


class SQLiteBackend(DatabaseBackend):

    driver_module = "sqlite3"
    list_tables_sql = "SELECT name FROM sqlite_master WHERE type = 'table'"

    def connect(self, dburi):
        # Ensure that multiple connections share the same data
        # https://sqlite.org/sharedcache.html
        conn = self.driver.connect(
            f"file:{dburi.database}?cache=shared",
            uri=True,
            detect_types=self.driver.PARSE_DECLTYPES,
        )
        conn.isolation_level = None
        return conn



from collections.abc import Mapping
from datetime import datetime
import time

from classic.migrations import utils
from classic.migrations.backends.base import DatabaseBackend

class PyMSSQLBackend(DatabaseBackend):
    driver_module = "pymssql"

    create_migration_table_sql = (
        "CREATE TABLE {0.migrations_schema_name_quoted}.{0.migration_table_quoted} ( "
        # sha256 hash of the migration id
        "migration_hash VARCHAR(64), "
        # The migration id (ie path basename without extension)
        "migration_id VARCHAR(255), "
        # When this id was applied
        "applied_at_utc DATETIME, "
        "PRIMARY KEY (migration_hash))"
    )
    insert_migration_table_from_log_table_sql = (
        "INSERT INTO {0.migrations_schema_name_quoted}.{0.migration_table_quoted} "
        "SELECT migration_hash, migration_id, created_at_utc "
        "FROM {0.migrations_schema_name_quoted}.{0.log_table_quoted}"
    )
    create_lock_table_sql = (
        "CREATE TABLE {0.migrations_schema_name_quoted}.{0.lock_table_quoted} ("
        "locked INT DEFAULT 1, "
        "ctime DATETIME,"
        "pid INT NOT NULL,"
        "PRIMARY KEY (locked))"
    )
    create_log_table_sql = (
        "CREATE TABLE {0.migrations_schema_name_quoted}.{0.log_table_quoted} ( "
        "id VARCHAR(36), "
        "migration_hash VARCHAR(64), "
        "migration_id VARCHAR(255), "
        "operation VARCHAR(10), "
        "username VARCHAR(255), "
        "hostname VARCHAR(255), "
        "comment VARCHAR(255), "
        "created_at_utc DATETIME, "
        "PRIMARY KEY (id))"
    )
    migrations_schema_exists_sql = "SELECT 1 FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = {0.migrations_schema_name_quoted};"

    def connect(self, dburi):
        return self.driver.connect(
            server=dburi.hostname,
            user=dburi.username,
            password=dburi.password,
            database=dburi.database,
            port=dburi.port,
        )

    def execute(self, sql, params=None):
        """
        Create a new cursor, execute a single statement and return the cursor
        object.

        :param sql: A single SQL statement, optionally with named parameters
                    (eg 'SELECT * FROM foo WHERE :bar IS NULL')
        :param params: A dictionary of parameters
        """
        if params and not isinstance(params, Mapping):
            raise TypeError("Expected dict or other mapping object")
        cursor = self.cursor()
        sql, params = utils.change_param_style(self.driver.paramstyle, sql, params)
        cursor.execute(sql, params)
        return cursor

    def begin(self):
        """
        Begin a new transaction
        """
        assert not self._in_transaction
        self._in_transaction = True

    def savepoint(self, id):
        """
        Create a new savepoint with the given id
        """
        pass

    def savepoint_release(self, id):
        """
        Release (commit) the savepoint with the given id
        """
        pass

    def savepoint_rollback(self, id):
        """
        Rollback the savepoint with the given id
        """
        self.connection.commit()

    def commit(self):
        try:
            self.connection.commit()
        except Exception as e:
            print(str(e))
        self._in_transaction = False

